#ifndef __EDIT_H
#define __EDIT_H

#ifdef MIDNIGHT

#ifdef HAVE_SLANG
#define HAVE_SYNTAXH 1
#endif

#    include <stdio.h>
#    include <stdarg.h>
#    include <sys/types.h>
#ifdef SOLARIS_SPARC
typedef uint32_t u_int32_t;
typedef uint8_t u_int8_t;
typedef uint16_t u_int16_t;
#endif

#    ifdef HAVE_UNISTD_H
#    	 include <unistd.h>
#    endif
#    include <string.h>
#    include "tty.h"
#    include <sys/stat.h>
#    include <errno.h>
     
#    ifdef HAVE_FCNTL_H
#        include <fcntl.h>
#    endif
     
#    include <stdlib.h>
#    include <malloc.h>

#else       /* ! MIDNIGHT */

// #    include "global.h"
#    include <stdio.h>
#    include <stdarg.h>
#    include <sys/types.h>
     
#    	 ifdef HAVE_UNISTD_H
#    	     include <unistd.h>
#    	 endif
     
#    include <my_string.h>
#    include <sys/stat.h>
     
#    ifdef HAVE_FCNTL_H
#    	 include <fcntl.h>
#    endif
     
#    include <stdlib.h>
#    include <stdarg.h>

#    if TIME_WITH_SYS_TIME
#    	 include <sys/time.h>
#    	 include <time.h>
#    else
#    	 if HAVE_SYS_TIME_H
#    	     include <sys/time.h>
#    	 else
#    	     include <time.h>
#    	 endif
#    endif
 
#    include "regex.h"

#endif

#    include "dlg.h"
#    include "widget.h"
#    include "color.h"
#    include "dialog.h"
#    include "mouse.h"
#    include "key.h"
#    include "wtools.h"		/* for QuickWidgets */
#    include "win.h"
#    include "menu.h"
#    include "regex.h"
#    define WANT_WIDGETS
     
#    define WIDGET_COMMAND (WIDGET_USER + 10)
// #    define N_menus 6

#define SEARCH_DIALOG_OPTION_NO_SCANF	1
#define SEARCH_DIALOG_OPTION_NO_REGEX	2
#define SEARCH_DIALOG_OPTION_NO_CASE	4
#define SEARCH_DIALOG_OPTION_BACKWARDS	8

#define SYNTAX_FILE "/.cedit/Syntax"
#define CLIP_FILE "/.cedit/cooledit.clip"
#define MACRO_FILE "/.cedit/cooledit.macros"
#define BLOCK_FILE "/.cedit/cooledit.block"
#define ERROR_FILE "/.cedit/cooledit.error"
#define TEMP_FILE "/.cedit/cooledit.temp"
#define SCRIPT_FILE "/.cedit/cooledit.script"
#define EDIT_DIR "/.cedit"
#define PROFILE_NAME     ".cedit/ini"

#define RC_FILE "cooledrc"
#define SYNTAX_DIR "syntax"
#define FILETYPES_FILE "filetypes"

#    define PATH_SEP '/'
#    define PATH_SEP_STR "/"
#    define PATH_ENV_SEP ':'



#ifndef MAXPATHLEN
#   define MC_MAXPATHLEN 4096
#else
#   define MC_MAXPATHLEN MAXPATHLEN
#endif



#define EDIT_KEY_EMULATION_NORMAL 0
#define EDIT_KEY_EMULATION_EMACS  1

#define REDRAW_LINE          (1 << 0)
#define REDRAW_LINE_ABOVE    (1 << 1)
#define REDRAW_LINE_BELOW    (1 << 2)
#define REDRAW_AFTER_CURSOR  (1 << 3)
#define REDRAW_BEFORE_CURSOR (1 << 4)
#define REDRAW_PAGE          (1 << 5)
#define REDRAW_IN_BOUNDS     (1 << 6)
#define REDRAW_CHAR_ONLY     (1 << 7)
#define REDRAW_COMPLETELY    (1 << 8)

#define MOD_ABNORMAL		(1 << 0)
#define MOD_UNDERLINED		(1 << 1)
#define MOD_BOLD		(1 << 2)
#define MOD_HIGHLIGHTED		(1 << 3)
#define MOD_MARKED		(1 << 4)
#define MOD_ITALIC		(1 << 5)
#define MOD_CURSOR		(1 << 6)
#define MOD_INVERSE		(1 << 7)

#    define EDIT_TEXT_HORIZONTAL_OFFSET 0
#    define EDIT_TEXT_VERTICAL_OFFSET 1
#    define FONT_OFFSET_X 0
#    define FONT_OFFSET_Y 0

#define EDIT_RIGHT_EXTREME option_edit_right_extreme
#define EDIT_LEFT_EXTREME option_edit_left_extreme
#define EDIT_TOP_EXTREME option_edit_top_extreme
#define EDIT_BOTTOM_EXTREME option_edit_bottom_extreme

#define MAX_MACRO_LENGTH 1024

/*there are a maximum of ... */
#define MAXBUFF 1024
/*... edit buffers, each of which is ... */
#define EDIT_BUF_SIZE 16384
/* ...bytes in size. */

/*x / EDIT_BUF_SIZE equals x >> ... */
#define S_EDIT_BUF_SIZE 14

/* x % EDIT_BUF_SIZE is equal to x && ... */
#define M_EDIT_BUF_SIZE 16383

#define SIZE_LIMIT (EDIT_BUF_SIZE * (MAXBUFF - 2))
/* Note a 16k stack is 64k of data and enough to hold (usually) around 10
   pages of undo info. */

/* undo stack */
#define START_STACK_SIZE 32

/*
#define INPUT_COMPLETE_FILENAMES         1
#define INPUT_COMPLETE_HOSTNAMES         2
#define INPUT_COMPLETE_COMMANDS          4
#define INPUT_COMPLETE_VARIABLES         8
#define INPUT_COMPLETE_USERNAMES        16
#define INPUT_COMPLETE_CD               32
*/

/*some codes that may be pushed onto or returned from the undo stack: */
#define CURS_LEFT 601
#define CURS_RIGHT 602
#define DELETE 603
#define BACKSPACE 604
#define STACK_BOTTOM 605
#define CURS_LEFT_LOTS 606
#define CURS_RIGHT_LOTS 607
#define MARK_1 1000
#define MARK_2 700000000
#define KEY_PRESS 1400000000

/*Tabs spaces: (sofar only HALF_TAB_SIZE is used: */
#define TAB_SIZE		option_tab_spacing
#define HALF_TAB_SIZE		((int) option_tab_spacing / 2)

struct macro {
    short command;
    short ch;
};

struct selection {
   unsigned char * text;
   int len;
};


#define RULE_CONTEXT		0x00FFF000UL
#define RULE_CONTEXT_SHIFT	12
#define RULE_WORD		0x00000FFFUL
#define RULE_WORD_SHIFT		0
#define RULE_ON_LEFT_BORDER	0x02000000UL
#define RULE_ON_RIGHT_BORDER	0x01000000UL

struct syntax_rule {
    unsigned short keyword;
    unsigned char end;
	unsigned char context;
	unsigned char _context;
	unsigned char border;
};
					

struct editor_widget {
    Widget widget;
#define from_here num_widget_lines
    int num_widget_lines;
    int num_widget_columns;

    int have_frame;

    char *filename;		/* Name of the file */
    char *dir;			/* current directory */

/* dynamic buffers and curser position for editor: */
    long curs1;			/*position of the cursor from the beginning of the file. */
    long curs2;			/*position from the end of the file */
    unsigned char *buffers1[MAXBUFF + 1];	/*all data up to curs1 */
    unsigned char *buffers2[MAXBUFF + 1];	/*all data from end of file down to curs2 */

/* search variables */
    long search_start;		/* First character to start searching from */
    int found_len;		/* Length of found string or 0 if none was found */
    long found_start;		/* the found word from a search - start position */

/* display information */
    long last_byte;		/* Last byte of file */
    long start_display;		/* First char displayed */
    long start_col;		/* First displayed column, negative */
    long max_column;		/* The maximum cursor position ever reached used to calc hori scroll bar */
    long curs_row;		/*row position of curser on the screen */
    long curs_col;		/*column position on screen */
    int force;			/* how much of the screen do we redraw? */
    unsigned char overwrite;
    unsigned char modified;	/*has the file been changed?: 1 if char inserted or
				   deleted at all since last load or save */
    int delete_file;			/* has the file been created in edit_load_file? Delete
			           it at end of editing when it hasn't been modified 
				   or saved */
    unsigned char highlight;
    long prev_col;		/*recent column position of the curser - used when moving
				   up or down past lines that are shorter than the current line */
    long curs_line;		/*line number of the cursor. */
    long start_line;		/*line nummber of the top of the page */

/* file info */
    long total_lines;		/*total lines in the file */
    long mark1;			/*position of highlight start */
    long mark2;			/*position of highlight end */
    int column1;			/*position of column highlight start */
    int column2;			/*position of column highlight end */
    long bracket;		/*position of a matching bracket */

/* undo stack and pointers */
    unsigned long stack_pointer;
    long *undo_stack;
    unsigned long stack_size;
    unsigned long stack_size_mask;
    unsigned long stack_bottom;
    struct stat stat;

/* syntax higlighting */

/* syntax higlighting */
    struct _syntax_marker *syntax_marker;
    struct context_rule **rules;
	long last_get_rule;
	struct syntax_rule rule;
	char *syntax_type;          /* description of syntax highlighting type being used */
	int explicit_syntax;        /* have we forced the syntax hi. type in spite of the filename? */
						 
    int to_here;		/* dummy marker */

/* file types */
    char *ft_name;
    char *ft_filename;
    char *ft_text;
    char *ft_command;
    char *ft_help;
    char *syntax_file;
    char *rc_file;


/* macro stuff */
    int macro_i;		/* -1 if not recording index to macro[] otherwise */
    struct macro macro[MAX_MACRO_LENGTH];

/* multifile */
    struct editor_widget * next;
    unsigned short fullscreen;
    unsigned short had_ysize;
    unsigned short had_pos;
    unsigned char running;
    
    int readonly;
};

typedef struct editor_widget WEdit;

/* util.c */
FILE *open_config_file(const char *file);
FILE *open_system_config_file(const char *file);
FILE *open_user_config_file(const char *file);
char *name_trunc(char *txt, int trunc_len);
char *strcasechr(const unsigned char *s, int c);
char *catstrs(const char *first, ...);
char *copy_strings(const char *first, ...);
char *concat_dir_and_file(const char *dir, const char *file);
char *working_dir();
int is_printable (int c);
int msglen (char *text, int *lines);
char * whitespace_lastindex(char *);
void whitespace_trim(char *);
int my_system (int flags, const char *shell, const char *command);
int block_signals (void);
int unblock_signals (void);
char *tilde_expand (char *directory);			    
int calc_buffer_size(WEdit * edit);

/* multifile.c */
void multifile_init(WEdit * we);
WEdit * multifile_new(char * file, int append);
void multifile_next(int);
void multifile_prev(int);
void multifile_clean_markers(WEdit * we);
int (*multifile_getcallback(void))(Dlg_head * h, WEdit * e, int msg, int par);
void multifile_resize(WEdit * we, int nsize);
void multifile_fullscreen(WEdit * we);
int multifile_quit(WEdit * we);
void multifile_cleanup(void);
int multifile_count(void);
void multifile_shutdown(void);
void multifile_select(WEdit * we);
void multifile_nofullscreen(void);
void multifile_switch(int n, int fs);
void multifile_switchto(WEdit *e, int fs);
void multifile_filebar_refresh(void);
WEdit * multifile_findfile(char * file);

/* unikey.c */
int unikey_load_mapping(FILE *fh);
void unikey_decode(u_int32_t x_state, int32_t x_key, int * command, int * charforins);
unsigned char ** unikey_create_help(void);
void unikey_dispose_help(unsigned char ** list);
u_int32_t unikey_translatecommand(unsigned char * cmd);

/* setup.c */
void read_setup(const char *file);
void save_setup(void);

/* pipes.c */
void handle_pipeoper(WEdit * we, int pid);
int register_pipeoper(char * name, int needs_block);

/* storeline.c */
int storeline_get(const char * filename);
void storeline_save(const char * filename, int line);


int edit_drop_hotkey_menu(WEdit * e, int key);
void edit_menu_cmd(WEdit * e);
int edit_init_menu_normal(void);
void edit_done_menu(void);
int edit_raw_key_query(char *heading, char *query, int cancel);
char *strcasechr(const unsigned char *s, int c);
int edit(char ** fa, int line);
int edit_translate_key(WEdit * edit, unsigned int x_keycode, long x_key, int x_state,
                                           int *cmd, int *ch);
int edit_get_byte(WEdit * edit, long byte_index);
char *edit_get_buffer_as_text(WEdit * edit);
int edit_load_file(WEdit * edit, const char *filename, const char *text,
                                   unsigned long text_size);
int edit_count_lines(WEdit * edit, long current, int upto);
long edit_move_forward(WEdit * edit, long current, int lines, long upto);
long edit_move_forward3(WEdit * edit, long current, int cols, long upto);
long edit_move_backward(WEdit * edit, long current, int lines);
void edit_scroll_screen_over_cursor(WEdit * edit);
void edit_scroll_upward(WEdit * edit, unsigned long i);
void edit_scroll_downward(WEdit * edit, int i);
void edit_scroll_right(WEdit * edit, int i);
void edit_scroll_left(WEdit * edit, int i);
int edit_get_col(WEdit * edit);
long edit_bol(WEdit * edit, long current);
long edit_eol(WEdit * edit, long current);
void edit_update_curs_row(WEdit * edit);
void edit_update_curs_col(WEdit * edit);

void edit_block_copy_cmd(WEdit * edit);
void edit_block_move_cmd(WEdit * edit);
int edit_block_delete_cmd(WEdit * edit);

int edit_delete(WEdit * edit);
void edit_insert(WEdit * edit, int c);
int edit_cursor_move(WEdit * edit, long increment);
void edit_push_action(WEdit * edit, long c, ...);
void edit_push_key_press(WEdit * edit);
void edit_insert_ahead(WEdit * edit, int c);
int edit_save_file(WEdit * edit, const char *filename);
int edit_save_cmd(WEdit * edit);
int edit_save_confirm_cmd(WEdit * edit);
int edit_save_as_cmd(WEdit * edit);
WEdit *edit_init(WEdit * edit, int lines, int columns, const char *filename,
                                 const char *text, const char *dir, unsigned long text_size);
int edit_clean(WEdit * edit);
int edit_renew(WEdit * edit);
int edit_new_cmd(WEdit * edit);
int edit_reload(WEdit * edit, const char *filename, const char *text, const char *dir,
                                unsigned long text_size);
int edit_load_cmd(WEdit * edit, int ignore_nonexistance);
int edit_load_new_cmd (WEdit * edit);
void edit_mark_cmd(WEdit * edit, int unmark);
void edit_set_markers(WEdit * edit, long m1, long m2, int c1, int c2);
void edit_push_markers(WEdit * edit);
int edit_quit_cmd(WEdit * edit, int);
void edit_replace_cmd(WEdit * edit, int again);
void edit_search_cmd(WEdit * edit, int again);
int edit_save_block_cmd(WEdit * edit);
int edit_insert_file_cmd(WEdit * edit);
int edit_insert_file(WEdit * edit, const char *filename);
void edit_block_process_cmd(WEdit * edit, const char *shell_cmd, int block);
char *catstrs(const char *first, ...);
void edit_refresh_cmd(WEdit * edit);
void edit_about_cmd ();
void edit_options_dialog();
void edit_wrap_cmd();
void edit_date_cmd(WEdit * edit);
void edit_goto_cmd(WEdit * edit);
int eval_marks(WEdit * edit, long *start_mark, long *end_mark);
void edit_status(WEdit * edit);
int edit_execute_command(WEdit * edit, int command, int char_for_insertion);
int edit_execute_key_command(WEdit * edit, int command, int char_for_insertion);
void edit_update_screen(WEdit * edit);
int edit_printf(WEdit * e, const char *fmt, ...);
int edit_print_string(WEdit * e, const char *s);
void edit_move_to_line(WEdit * e, long line);
void edit_move_display(WEdit * e, long line);
void edit_word_wrap(WEdit * edit);
unsigned char *edit_get_block(WEdit * edit, long start, long finish, int *l);
int edit_sort_cmd(WEdit * edit);
void edit_execute(WEdit * edit, char *cmd);
void edit_help_cmd(WEdit * edit);
void edit_left_word_move(WEdit * edit);
void edit_right_word_move(WEdit * edit);
void edit_get_selection(WEdit * edit);
void edit_drop_menu_cmd (WEdit * e, int which);

int edit_save_macro_cmd(WEdit * edit, struct macro macro[], int n);
int edit_load_macro_cmd(WEdit * edit, struct macro macro[], int *n, int k);
void edit_delete_macro_cmd(WEdit * edit);

int edit_copy_to_X_buf_cmd(WEdit * edit);
int edit_cut_to_X_buf_cmd(WEdit * edit);
void edit_paste_from_X_buf_cmd(WEdit * edit);

void edit_paste_from_history(WEdit * edit);

void edit_split_filename(WEdit * edit, char *name);

char *edit_expandformat(WEdit * edit, char *format, int maxlen);




int edit_drop_hotkey_menu (WEdit * e, int key);
void edit_menu_cmd (WEdit * e);
void edit_init_menu_emacs (void);
int edit_init_menu_normal (void);
void edit_done_menu (void);
int edit_raw_key_query (char *heading, char *query, int cancel);
char *strcasechr (const unsigned char *s, int c);
int edit_translate_key (WEdit * edit, unsigned int x_keycode, long x_key, int x_state, int *cmd, int *ch);

int edit_get_byte (WEdit * edit, long byte_index);
char *edit_get_buffer_as_text (WEdit * edit);
int edit_load_file (WEdit * edit, const char *filename, const char *text, unsigned long text_size);
int edit_count_lines (WEdit * edit, long current, int upto);
long edit_move_forward (WEdit * edit, long current, int lines, long upto);
long edit_move_forward3 (WEdit * edit, long current, int cols, long upto);
long edit_move_backward (WEdit * edit, long current, int lines);
void edit_scroll_screen_over_cursor (WEdit * edit);
void edit_render_keypress (WEdit * edit);
void edit_scroll_upward (WEdit * edit, unsigned long i);
void edit_scroll_downward (WEdit * edit, int i);
void edit_scroll_right (WEdit * edit, int i);
void edit_scroll_left (WEdit * edit, int i);
int edit_get_col (WEdit * edit);
long edit_bol (WEdit * edit, long current);
long edit_eol (WEdit * edit, long current);
void edit_update_curs_row (WEdit * edit);
void edit_update_curs_col (WEdit * edit);

void edit_block_copy_cmd (WEdit * edit);
void edit_block_move_cmd (WEdit * edit);
int edit_block_delete_cmd (WEdit * edit);

int edit_delete (WEdit * edit);
void edit_insert (WEdit * edit, int c);
int edit_cursor_move (WEdit * edit, long increment);
void edit_push_action (WEdit * edit, long c,...);
void edit_push_key_press (WEdit * edit);
void edit_insert_ahead (WEdit * edit, int c);
int edit_save_file (WEdit * edit, const char *filename);
int edit_save_cmd (WEdit * edit);
int edit_save_confirm_cmd (WEdit * edit);
int edit_save_as_cmd (WEdit * edit);
WEdit *edit_init (WEdit * edit, int lines, int columns, const char *filename, const char *text, const char *dir, unsigned long text_size);
int edit_clean (WEdit * edit);
int edit_renew (WEdit * edit);
int edit_new_cmd (WEdit * edit);
int edit_reload (WEdit * edit, const char *filename, const char *text, const char *dir, unsigned long text_size);
void edit_mark_cmd (WEdit * edit, int unmark);
void edit_set_markers (WEdit * edit, long m1, long m2, int c1, int c2);
void edit_push_markers (WEdit * edit);
void edit_replace_cmd (WEdit * edit, int again);
void edit_search_cmd (WEdit * edit, int again);
int edit_save_block_cmd (WEdit * edit);
int edit_insert_file_cmd (WEdit * edit);
int edit_insert_file (WEdit * edit, const char *filename);
void edit_block_process_cmd (WEdit * edit, const char *shell_cmd, int block);
char *catstrs (const char *first,...);
void edit_refresh_cmd (WEdit * edit);
void edit_date_cmd (WEdit * edit);
void edit_goto_cmd (WEdit * edit);
int eval_marks (WEdit * edit, long *start_mark, long *end_mark);
void edit_status (WEdit * edit);
int edit_execute_command (WEdit * edit, int command, int char_for_insertion);
int edit_execute_key_command (WEdit * edit, int command, int char_for_insertion);
void edit_update_screen (WEdit * edit);
int edit_printf (WEdit * e, const char *fmt,...);
int edit_print_string (WEdit * e, const char *s);
void edit_move_to_line (WEdit * e, long line);
void edit_move_display (WEdit * e, long line);
void edit_word_wrap (WEdit * edit);
unsigned char *edit_get_block (WEdit * edit, long start, long finish, int *l);
int edit_sort_cmd (WEdit * edit);
void edit_help_cmd (WEdit * edit);
void edit_left_word_move (WEdit * edit);
void edit_right_word_move (WEdit * edit);
void edit_get_selection (WEdit * edit);

int edit_save_macro_cmd (WEdit * edit, struct macro macro[], int n);
int edit_load_macro_cmd (WEdit * edit, struct macro macro[], int *n, int k);
void edit_delete_macro_cmd (WEdit * edit);

int edit_copy_to_X_buf_cmd (WEdit * edit);
int edit_cut_to_X_buf_cmd (WEdit * edit);
void edit_paste_from_X_buf_cmd (WEdit * edit);

void edit_paste_from_history (WEdit *edit);

void edit_split_filename (WEdit * edit, char *name);

void edit_labels (WEdit * edit);
long get_key_state ();



#define CWidget Widget

void edit_set_syntax_change_callback (void (*callback) (CWidget *));
void edit_load_syntax (WEdit * edit, char **names, char *type);
void edit_free_syntax_rules (WEdit * edit);
void edit_get_syntax_color (WEdit * edit, long byte_index, int *color);

/* put OS2/NT/WIN95 defines here */

#    ifdef OS2_NT
#    	 define MY_O_TEXT O_TEXT
#    else
#    	 define MY_O_TEXT 0
#    endif

#    define FONT_PIX_PER_LINE 1
#    define FONT_MEAN_WIDTH 1
     
#    define get_sys_error(s) (s)
#    define itoa MY_itoa

#    define edit_get_load_file(d,f,h) input_dialog (h, _(" Enter file name: "), f)
#    define edit_get_file(d,f,h) file_input_dialog (h, _(" Enter file name: "), f)
#    define edit_get_save_file(d,f,h) input_dialog (h, _(" Enter file name: "), f)
#    define CMalloc(x) malloc(x)
     
#    define set_error_msg(s) edit_init_error_msg = strdup(s)

#    ifdef _EDIT_C

#         define edit_error_dialog(h,s) set_error_msg(s)
char *edit_init_error_msg = NULL;

#    else				/* ! _EDIT_C */

#    define edit_error_dialog(h,s) query_dialog (h, s, 0, 1, _("&Dismiss"))
#    define edit_message_dialog(h,s) query_dialog (h, s, 0, 1, _("&Ok"))
extern char *edit_init_error_msg;

#    endif				/* ! _EDIT_C */


#    define get_error_msg(s) edit_init_error_msg
#    define edit_query_dialog2(h,t,a,b) query_dialog(h,t,0,2,a,b)
#    define edit_query_dialog3(h,t,a,b,c) query_dialog(h,t,0,3,a,b,c)
#    define edit_query_dialog4(h,t,a,b,c,d) query_dialog(h,t,0,4,a,b,c,d)


void *do_xmalloc(int);
#define xmalloc(a,b) do_xmalloc (a)

extern char *home_dir;
extern char *conf_dir;


#    ifdef ENABLE_NLS
#        include <libintl.h>
#        define _(String) gettext (String)
#        ifdef gettext_noop
#            define N_(String) gettext_noop (String)
#        else
#            define N_(String) (String)
#        endif
#    else
                 /* Stubs that do something close enough.  */
#        define textdomain(String) (String)
#        define gettext(String) (String)
#        define dgettext(Domain,Message) (Message)
#        define dcgettext(Domain,Message,Type) (Message)
#        define bindtextdomain(Domain,Directory) (Domain)
#        define _(String) (String)
#        define N_(String) (String)
#    endif
		 
		 
#ifdef min
#undef min
#endif
		 
#ifdef max
#undef max
#endif
 
#define min(x,y) (x>y ? y: x)
#define max(x,y) (x>y ? x: y)

#define ESC_CHAR '\033'
#define ESC_STR  "\033"

#ifdef USE_BSD_CURSES
#   define xgetch x_getch
#else
#   define xgetch getch
#endif
 

#ifndef MAX_PATH_LEN
#ifdef PATH_MAX
#define MAX_PATH_LEN PATH_MAX
#else
#define MAX_PATH_LEN 1024
#endif
#endif

#define NUM_SELECTION_HISTORY 32

#ifdef _EDIT_C

struct selection selection =
{0, 0};
int current_selection = 0;
/* Note: selection.text = selection_history[current_selection].text */
struct selection selection_history[NUM_SELECTION_HISTORY] =
{
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
    {0, 0},
};

/*
   what editor are we going to emulate? one of EDIT_KEY_EMULATION_NORMAL
   or EDIT_KEY_EMULATION_EMACS
 */
int edit_key_emulation = EDIT_KEY_EMULATION_NORMAL;

extern int option_syntax_highlighting;

int option_word_wrap_line_length = 72;
int option_typewriter_wrap = 0;
int option_auto_para_formatting = 0;
int option_international_characters = 0;
int option_tab_spacing = 8;
int option_fill_tabs_with_spaces = 0;
int option_return_does_auto_indent = 1;
int option_backspace_through_tabs = 0;
int option_fake_half_tabs = 0;
int option_save_mode = 0;
int option_backup_ext_int = -1;
int option_find_bracket = 1;
int option_max_undo = 8192;

int option_eight_bit_clean = 1;
int option_storeline = 1;

int option_editor_fg_normal = 26;
int option_editor_fg_bold = 8;
int option_editor_fg_italic = 10;

int option_edit_right_extreme = 0;
int option_edit_left_extreme = 0;
int option_edit_top_extreme = 0;
int option_edit_bottom_extreme = 0;

int option_editor_bg_normal = 1;
int option_editor_bg_abnormal = 0;
int option_editor_bg_marked = 2;
int option_editor_bg_marked_abnormal = 9;
int option_editor_bg_highlighted = 12;
int option_editor_fg_cursor = 18;

int option_auto_fullscreen = 1;

char *option_whole_chars_search = "0123456789abcdefghijklmnopqrstuvwxyz_";
char *option_whole_chars_move = "0123456789abcdefghijklmnopqrstuvwxyz_; ,[](){}";
char *option_backup_ext = "~";

#else				/* ! _EDIT_C */

extern struct selection selection;
extern struct selection selection_history[];
extern int current_selection;

/*
   what editor are we going to emulate? one of EDIT_KEY_EMULATION_NORMAL
   or EDIT_KEY_EMULATION_EMACS
 */
extern int edit_key_emulation;

extern int option_word_wrap_line_length;
extern int option_typewriter_wrap;
extern int option_auto_para_formatting;
extern int option_international_characters;
extern int option_tab_spacing;
extern int option_fill_tabs_with_spaces;
extern int option_return_does_auto_indent;
extern int option_backspace_through_tabs;
extern int option_fake_half_tabs;
extern int option_save_mode;
extern int option_backup_ext_int;
extern int option_find_bracket;
extern int option_max_undo;

extern int option_editor_fg_normal;
extern int option_editor_fg_bold;
extern int option_editor_fg_italic;

extern int option_edit_right_extreme;
extern int option_edit_left_extreme;
extern int option_edit_top_extreme;
extern int option_edit_bottom_extreme;

extern int option_editor_bg_normal;
extern int option_editor_bg_abnormal;
extern int option_editor_bg_marked;
extern int option_editor_bg_marked_abnormal;
extern int option_editor_bg_highlighted;
extern int option_editor_fg_cursor;

extern int option_eight_bit_clean;
extern int option_storeline;
extern int option_auto_fullscreen;

extern char *option_whole_chars_search;
extern char *option_whole_chars_move;
extern char *option_backup_ext;

extern int edit_confirm_save;

#endif				/* ! _EDIT_C */
#endif 				/* __EDIT_H */
