/* editor menu definitions and initialisation

   Copyright (C) 1996, 2003 the Free Software Foundation

   Authors: 1996, 1997 Paul Sheer
                  2003 Mateusz Golicz <mteg@jaszczur.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#include <ctype.h>
#include "config.h"
#include "edit.h"

#include "editcmddef.h"

extern int edit_key_emulation;
extern WEdit *wedit_current;
extern WButtonBar *edit_bar;
extern Dlg_head *edit_dlg;
extern WMenu *edit_menubar;
int N_menus = 0;
#define isblank(c) ((c) == ' ' || (c) == '\t')
#undef edit_message_dialog
#define edit_message_dialog(w,x,y,h,s) query_dialog (h, s, 0, 1, "&Ok")
#define CFocus(x) 

static void menu_default_menus(void);

static void menu_cmd (int i)
{
    send_message (wedit_current->widget.parent, (Widget *) wedit_current, WIDGET_COMMAND, i);
}

/*static void menu_key (int i)
{
    send_message (wedit_current->widget.parent, (Widget *) wedit_current, WIDGET_KEY, i);
}*/

void edit_wrap_cmd ()
{
    char *f;
    char s[12];
    sprintf (s, "%d", option_word_wrap_line_length);
    f = input_dialog (_(" Word wrap "), 
/* Not essential to translate */
    _(" Enter line length, 0 for off: "), s);
    if (f) {
	if (*f) {
	    option_word_wrap_line_length = atoi (f);
	    free (f);
	}
    }
}

void edit_about_cmd ()
{
    edit_message_dialog (wedit_current->mainid, 20, 20, " About ",
		      "\n"
		      "                 Cooledit  v2.1\n"
		      "\n"
		      " Copyright (C) 1996 - 2003 the Free Software Foundation\n"
		      "\n"
		      "       A user friendly text editor written\n"
		      "           for the Midnight Commander.\n"
	);
}
static void menu_menu_cmd(int cmd)	{ menu_cmd (cmd); }

extern void menu_save_mode_cmd (void);

callfn dispatch_command_callback (u_int32_t cmd)
{
	return(menu_menu_cmd);
}

Menu *EditMenuBar = NULL;

int edit_init_menu_normal (void)
{
	FILE * fh;
	unsigned char * line;
	unsigned char buffer[256];
	
	unsigned char * lastname = NULL;
	menu_entry * lastrecords = NULL;
	int lastcount = 0;
	
	int lineno = 0, hotkey, i;
	
	fh = open_config_file("menu");

	if(!fh)
	{
		menu_default_menus();
		return(1);
	}

        while((line = fgets(buffer, 255, fh)))
        {
                unsigned char *r;
                u_int32_t action;

		lineno ++;
                while(isblank(*line)) line++;


                if((r = index(line, '\n'))) *r = 0;
                if((r = index(line, '\r'))) *r = 0;
                if(*line == '#') continue;
		if(!strncmp(line, "menu ", 5))
		{
			if(lastrecords)
			{
				EditMenuBar = (Menu*) realloc(EditMenuBar, (N_menus + 1) * sizeof(Menu));
				EditMenuBar[N_menus++] = create_menu (lastname, lastrecords, lastcount);
			}
			lastrecords = NULL;
			lastname = strdup(line + 5);
			lastcount = 0;
			continue;
		}
		if(!strcmp(line, "sep"))
		{
			if(!lastname)
			{
				fprintf(stderr, "Error parsing menu file, line %d. Separator outside a menu.\n", lineno);
				return(0);
			}
			lastrecords = (menu_entry*) realloc(lastrecords, sizeof(menu_entry) * (lastcount + 1));
			lastrecords[lastcount].first_letter = ' ';
			lastrecords[lastcount].text = strdup("");
			lastrecords[lastcount].hot_key = 0;
			lastrecords[lastcount].call_back = 0;
			lastcount++;
			continue;
		}
                if(!(r = whitespace_lastindex(line))) continue;
		if(!lastname)
		{
			fprintf(stderr, "Error parsing menu file, line %d. Menu item outside a menu.\n", lineno);
			return(0);
		}
		
		whitespace_trim(line);
		action = unikey_translatecommand(line);
		if(!action)
		{
			fprintf(stderr, "Error parsing menu file, line %d. No such action '%s'\n", lineno, line);
			return(0);
		}
		if(!strlen(r))
		{
			fprintf(stderr, "Error parsing menu file, line %d. Syntax error.\n", lineno);
			return(0);
		}
		for(hotkey = 0, i = 1; r[i]; i++)
			if(r[i - 1] == '&') { hotkey = r[i]; break; }

		lastrecords = (menu_entry*) realloc(lastrecords, sizeof(menu_entry) * (lastcount + 1));
		lastrecords[lastcount].first_letter = ' ';
		lastrecords[lastcount].text = strdup(r);
		lastrecords[lastcount].hot_key = hotkey;
		lastrecords[lastcount].call_back = menu_menu_cmd;
		lastrecords[lastcount].userdata = action;
		if(!lastrecords[lastcount].call_back)
		{
			fprintf(stderr, "Error parsing menu file, line %d. Command '%s' not supported in menu.\n", lineno, line);
			return(0);
		}
		
		
		lastcount++;		
	}
	if(lastrecords)
	{
		EditMenuBar = (Menu*) realloc(EditMenuBar, (N_menus + 1) * sizeof(Menu));
		EditMenuBar[N_menus++] = create_menu (lastname, lastrecords, lastcount);
	}


	if(N_menus == 0)
	{
		fprintf(stderr, "No menus in fact defined!\n");
		return(0);
	}

	return(1);
}


void edit_done_menu (void)
{
    int i;
    for (i = 0; i < N_menus; i++)
	destroy_menu (EditMenuBar[i]);
}


void edit_drop_menu_cmd (WEdit * e, int which)
{
//    fprintf(stderr, "MENU ON\n");
    if (edit_menubar->active)
	return;
    edit_menubar->active = 1;
    edit_menubar->dropped = 0;
    edit_menubar->previous_selection = which >= 0 ? which : dlg_item_number (edit_dlg);
    if (which >= 0)
	edit_menubar->selected = which;
    dlg_select_widget (edit_dlg, edit_menubar);
}


void edit_menu_cmd (WEdit * e)
{
    edit_drop_menu_cmd (e, -1);
}


int edit_drop_hotkey_menu (WEdit * e, int key)
{
/*    switch (key) {
    case ALT ('f'):
	if (edit_key_emulation == EDIT_KEY_EMULATION_EMACS)
	    return 0;
	m = 0;
	break;
    case ALT ('e'):
	m = 1;
	break;
    case ALT ('s'):
	m = 2;
	break;
    case ALT ('c'):
	m = 3;
	break;
    case ALT ('o'):
	m = 4;
	break;
    default:
	return 0;
    }*/

//    edit_drop_menu_cmd (e, m);
    return 0;
}


/* defaults */
static menu_entry FileMenu[] =
{
    {' ', N_("&Open/load...     C-o"), 'O', menu_menu_cmd, CK_Load},
    {' ', N_("Open in new &wind F11"), 'W', menu_menu_cmd, CK_Load_NewWindow},
    {' ', N_("&Clear            C-n"), 'N', menu_menu_cmd, CK_New},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Save              F2"), 'S', menu_menu_cmd, CK_Save},
    {' ', N_("save &As...       F12"), 'A', menu_menu_cmd, CK_Save_As},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Insert file...   F15"), 'I', menu_menu_cmd, CK_Insert_File},
    {' ', N_("copy to &File...  C-f"), 'F', menu_menu_cmd, CK_Save_Block},
    {' ', "", ' ', 0, 0},
    {' ', N_("a&Bout...            "), 'B', edit_about_cmd, CK_About},
    {' ', N_("&Keys...            "), 'K', menu_menu_cmd, CK_Help},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Close            F10"), 'C', menu_menu_cmd, CK_Exit},
    {' ', N_("&Quit        Shft F10"), 'Q', menu_menu_cmd, CK_Shutdown}
};

static menu_entry EditMenu[] =
{
    {' ', N_("&Toggle Mark       F3"), 'T', menu_menu_cmd, CK_Mark},
    {' ', "", ' ', 0, 0},
    {' ', N_("toggle &Ins/overw Ins"), 'I', menu_menu_cmd, CK_Toggle_Insert},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Copy              F5"), 'C', menu_menu_cmd, CK_Copy},
    {' ', N_("&Move              F6"), 'M', menu_menu_cmd, CK_Move},
    {' ', N_("&Delete            F8"), 'D', menu_menu_cmd, CK_Remove},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Undo             C-u"), 'U', menu_menu_cmd, CK_Undo},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Beginning     C-PgUp"), 'B', menu_menu_cmd, CK_Beginning_Of_Text},
    {' ', N_("&End           C-PgDn"), 'E', menu_menu_cmd, CK_End_Of_Text}
};

static menu_entry SearReplMenu[] =
{
    {' ', N_("&Search...         F7"), 'S', menu_menu_cmd, CK_Find},
    {' ', N_("search &Again     F17"), 'A', menu_menu_cmd, CK_Find_Again},
    {' ', N_("&Replace...        F4"), 'R', menu_menu_cmd, CK_Replace}
};

static menu_entry CmdMenu[] =
{
    {' ', N_("&Goto line...            M-l"), 'G', menu_menu_cmd, CK_Goto},
    {' ', "", ' ', 0, 0},
    {' ', N_("insert &Literal...       C-q"), 'L', menu_menu_cmd, CK_InsertLiteral},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Refresh screen          C-l"), 'R', menu_menu_cmd, CK_Refresh},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Start record macro      C-r"), 'S', menu_menu_cmd, CK_Begin_Record_Macro},
    {' ', N_("&Finish record macro...  C-r"), 'F', menu_menu_cmd, CK_End_Record_Macro},
    {' ', N_("&Execute macro...   C-a, KEY"), 'E', menu_menu_cmd, CK_Execute_Macro},
    {' ', N_("delete macr&O...            "), 'O', menu_menu_cmd, CK_Delete_Macro},
    {' ', "", ' ', 0, 0},
    {' ', N_("insert &Date/time           "), 'D', menu_menu_cmd, CK_Date},
    {' ', "", ' ', 0, 0},
    {' ', N_("format p&Aragraph        M-p"), 'A', menu_menu_cmd, CK_Paragraph_Format},
    {' ', N_("sor&T...                 M-t"), 'T', menu_menu_cmd, CK_Sort},
    {' ', N_("&Mail...                    "), 'M', menu_menu_cmd, CK_Mail}
};

static menu_entry OptMenu[] =
{
    {' ', N_("&General...  "), 'G', menu_menu_cmd, CK_Options},
    {' ', N_("&Save mode..."), 'S', menu_menu_cmd, CK_Options_Save}
};

static menu_entry WindowMenu[] =
{
    {' ', N_("&New window        F1"), 'N', menu_menu_cmd, CK_NewWindow},
    {' ', N_("Spl&it window        "), 'I', menu_menu_cmd, CK_NewWindow_Split},
    {' ', "", ' ', 0, 0},
    {' ', N_("Ne&xt window    M Tab"), 'X', menu_menu_cmd, CK_Toggle},
    {' ', N_("&Prev window         "), 'P', menu_menu_cmd, CK_ToggleBackwards},
    {' ', "", ' ', 0, 0},
    {' ', N_("&Enlarge window   M z"), 'E', menu_menu_cmd, CK_SizePlus},
    {' ', N_("&Shrink window    M a"), 'S', menu_menu_cmd, CK_SizeMinus},
    {' ', N_("&Fullscreen   Alt Ent"), 'F', menu_menu_cmd, CK_Fullscreen},
};

#define menu_entries(x) sizeof(x)/sizeof(menu_entry)

static void menu_default_menus(void)
{
	N_menus = 6;
	EditMenuBar = (Menu*) malloc(N_menus * sizeof(Menu));
	EditMenuBar[0] = create_menu (strdup("File"), FileMenu, menu_entries(FileMenu));
	EditMenuBar[1] = create_menu (strdup("Edit"), EditMenu, menu_entries(EditMenu));
	EditMenuBar[2] = create_menu (strdup("S / R"), SearReplMenu, menu_entries(SearReplMenu));
	EditMenuBar[3] = create_menu (strdup("Command"), CmdMenu, menu_entries(CmdMenu));
	EditMenuBar[4] = create_menu (strdup("Window"), WindowMenu, menu_entries(WindowMenu));
	EditMenuBar[5] = create_menu (strdup("Options"), OptMenu, menu_entries(OptMenu));
}
