/* editor initialisation and callback handler.

   Copyright (C) 1996, 1997, 2003 the Free Software Foundation

   Authors: 1996, 1997 Paul Sheer
                  2003 Mateusz Golicz <mteg@jaszczur.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include "config.h"
#include "edit.h"
#include "widget.h"

char *tgetstr(const char *id, char **area);
	   
char * home_dir, * conf_dir;

WEdit *wedit;
WButtonBar *edit_bar;
Dlg_head *edit_dlg;
WMenu *edit_menubar;

int column_highlighting = 0;

static int edit_mode_callback (struct Dlg_head *h, int id, int msg)
{
    return 0;
}

int edit_event (WEdit * edit, Gpm_Event * event, int *result)
{
    *result = MOU_NORMAL;
    edit_update_curs_row (edit);
    edit_update_curs_col (edit);
    if (event->type & (GPM_DOWN | GPM_DRAG | GPM_UP)) {
	if (event->y > 1 && event->x > 0
	    && event->x <= edit->num_widget_columns
	    && event->y <= edit->num_widget_lines + 1) {
	    if (edit->mark2 != -1 && event->type & (GPM_UP | GPM_DRAG))
		return 1;	/* a lone up mustn't do anything */
	    if (event->type & (GPM_DOWN | GPM_UP))
		edit_push_key_press (edit);
	    edit_cursor_move (edit, edit_bol (edit, edit->curs1) - edit->curs1);
	    if (--event->y > (edit->curs_row + 1))
		edit_cursor_move (edit,
				  edit_move_forward (edit, edit->curs1, event->y - (edit->curs_row + 1), 0)
				  - edit->curs1);
	    if (event->y < (edit->curs_row + 1))
		edit_cursor_move (edit,
				  +edit_move_backward (edit, edit->curs1, (edit->curs_row + 1) - event->y)
				  - edit->curs1);
	    edit_cursor_move (edit, (int) edit_move_forward3 (edit, edit->curs1,
		       event->x - edit->start_col - 1, 0) - edit->curs1);
	    edit->prev_col = edit_get_col (edit);
	    if (event->type & GPM_DOWN) {
		edit_mark_cmd (edit, 1);	/* reset */
		edit->highlight = 0;
	    }
	    if (!(event->type & GPM_DRAG))
		edit_mark_cmd (edit, 0);
	    edit->force |= REDRAW_COMPLETELY;
	    edit_update_curs_row (edit);
	    edit_update_curs_col (edit);
	    edit_update_screen (edit);
	    multifile_select(edit);
	    return 1;
	}
    }
    return 0;
}



int menubar_event (Gpm_Event * event, WMenu * menubar);		/* menu.c */

int edit_mouse_event (Gpm_Event * event, void *x)
{
    int result;
    if (edit_event ((WEdit *) x, event, &result))
	return result;
    else
	return menubar_event (event, edit_menubar);
}

extern Menu * EditMenuBar;
extern int N_menus;

unsigned char * fetchline(unsigned char * name, int *line)
{
	int fd;
	unsigned char * colon;

	if(!name)
	{
		*line = 0;
		return(NULL);
	}
	
	if((fd = open(name, O_RDONLY)) != -1)
		close(fd);
	else
	{
		if((colon = rindex(name, ':')))
		{
			if(*(colon + 1))
			{
				unsigned char *nname = strdup(name);
				colon = rindex(nname, ':');
				*colon = 0;
				if((fd = open(nname, O_RDONLY)) == -1)
					free(nname);
				else
				{
					char * endptr;
					long int n;
				
					close(fd);
					n = strtol(colon + 1, &endptr, 0);
					if(*endptr || n < 1)
						free(nname);
					else
					{
						*line = n;
						return(nname);
					}
				}
			}
		}
	}
	*line = 0;
	return(strdup(name));
}

int edit (char **filearray, int line)
{
    static int made_directory = 0;
    int framed = 0, l, filenamelineno;
    int midnight_colors[4];
    char *text = 0;
	char *_file = filearray[0];

    read_setup(NULL);

    if (option_backup_ext_int != -1) {
	option_backup_ext = malloc (sizeof(int) + 1);
	option_backup_ext[sizeof(int)] = '\0';
	memcpy (option_backup_ext, (char *) &option_backup_ext_int, sizeof (int));
    }

    if (!made_directory) {
	mkdir (catstrs (home_dir, EDIT_DIR, 0), 0700);
	made_directory = 1;
    }
    if (_file) {
	if (!(*_file)) {
	    _file = 0;
	    text = "";
	}
    } else
	text = "";

	_file = fetchline(_file, &filenamelineno);

    if (!(wedit = edit_init (NULL, LINES - 2, COLS, _file, text, "", 0))) {
	message (1, _(" Error "), get_error_msg (""));
	return 0;
    }
    wedit->macro_i = -1;

    /* Create a new dialog and add it widgets to it */
    edit_dlg = create_dlg (0, 0, LINES, COLS, midnight_colors,
			   edit_mode_callback, "[Internal File Editor]",
			   "edit",
			   DLG_NONE);

    edit_dlg->raw = 1;		/*so that tab = '\t' key works */

    init_widget (&(wedit->widget), 0, 0, LINES - 1, COLS,
		 (callback_fn) multifile_getcallback(),
		 (destroy_fn) edit_clean,
		 (mouse_h) edit_mouse_event, 0);

    widget_want_cursor (wedit->widget, 1);

    edit_bar = buttonbar_new (1);

//    if (!framed) {
//	switch (edit_key_emulation) {
//	case EDIT_KEY_EMULATION_NORMAL:
//	    break;
//	case EDIT_KEY_EMULATION_EMACS:
//	    edit_init_menu_emacs ();	/* editmenu.c */
//	    break;
//	}
	edit_menubar = menubar_new (0, 0, COLS, EditMenuBar, N_menus);
//    }
    add_widget (edit_dlg, wedit);

    if (!framed)
	add_widget (edit_dlg, edit_menubar);

    add_widget (edit_dlg, edit_bar);

	if(filenamelineno)
		line = filenamelineno;
	else
    	line = storeline_get(_file);
	
	
    edit_move_display (wedit, ((l = line - 1 - (wedit->num_widget_lines / 2)) > 0) ? l : 0);
    edit_move_to_line (wedit, line - 1);



    multifile_init(wedit);

	if(filearray[0])
	{
		int i;
		for(i = 1; filearray[i]; i++)
		{
			WEdit * new;
			unsigned char * fn;
			
			fn = fetchline(filearray[i], &filenamelineno);
			new = multifile_new(fn, 1);
			if(filenamelineno && new)
			{
				edit_move_display (new, ((l = filenamelineno - 1 - (new->num_widget_lines / 2)) > 0) ? l : 0);
				edit_move_to_line (new, filenamelineno - 1);			
			}
			/* free fn ? */
		}
	}
	
	
    while(multifile_count() > 0)
    {
   	 run_dlg (edit_dlg);
    	 multifile_cleanup();
    }

    if (!framed)
	edit_done_menu ();	/* editmenu.c */

//    destroy_dlg (edit_dlg);
/* actually the list of widgets is quite damaged at this point, so we're not going to free resources */
    return 1;
}

static void edit_my_define (Dlg_head * h, int idx, char *text,
			    void (*fn) (WEdit *), WEdit * edit)
{
    define_label_data (h, (Widget *) edit, idx, text, (buttonbarfn) fn, edit);
}


void cmd_F1 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (1));
}

void cmd_F2 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (2));
}

void cmd_F3 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (3));
}

void cmd_F4 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (4));
}

void cmd_F5 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (5));
}

void cmd_F6 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (6));
}

void cmd_F7 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (7));
}

void cmd_F8 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (8));
}

void cmd_F9 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (9));
}

void cmd_F10 (WEdit * edit)
{
    send_message (edit->widget.parent, (Widget *) edit, WIDGET_KEY, KEY_F (10));
}

void edit_labels (WEdit * edit)
{
    Dlg_head *h = edit->widget.parent;

    edit_my_define (h, 1, _("New"), cmd_F1, edit);
    edit_my_define (h, 2, _("Save"), cmd_F2, edit);
    edit_my_define (h, 3, _("Mark"), cmd_F3, edit);
    edit_my_define (h, 4, _("Replac"), cmd_F4, edit);
    edit_my_define (h, 5, _("Copy"), cmd_F5, edit);
    edit_my_define (h, 6, _("Move"), cmd_F6, edit);
    edit_my_define (h, 7, _("Search"), cmd_F7, edit);
    edit_my_define (h, 8, _("Delete"), cmd_F8, edit);
    if (!edit->have_frame)
	edit_my_define (h, 9, _("PullDn"), edit_menu_cmd, edit);
    edit_my_define (h, 10, _("Close"), cmd_F10, edit);

    redraw_labels (h, (Widget *) edit);
}


long get_key_state ()
{
    return (long) get_modifier ();
}

void edit_adjust_size (Dlg_head * h)
{
#ifdef RESIZABLE_MENUBAR
	menubar_arrange(edit_menubar);
#endif
}

void edit_update_screen (WEdit * e)
{
    if(!e->widget.lines) return; /* hidden (somebody other is now fullscreen ) */

    edit_scroll_screen_over_cursor (e);

    edit_update_curs_col (e);
    edit_status (e);

/* pop all events for this window for internal handling */

    if (!is_idle ()) {
	e->force |= REDRAW_PAGE;
	return;
    }
    if (e->force & REDRAW_COMPLETELY)
	e->force |= REDRAW_PAGE;
    edit_render_keypress (e);
}



int main(int argc, char *argv[])
{
    unsigned char * clstr;
	
    home_dir = getenv("HOME");

    if (!home_dir)
    	home_dir = PATH_SEP_STR;

    conf_dir = strdup(concat_dir_and_file(home_dir, EDIT_DIR));

	if(!unikey_load_mapping(open_config_file("keymap"))) return(1);
	if(!edit_init_menu_normal ()) return(1);	/* editmenu.c */


    initscr();
	clstr = tgetstr("ti", NULL);
	if(clstr) putp(clstr);
	
	init_key();
	
	
    mc_raw_mode();
//        cbreak();
    noecho();
    keypad(stdscr, TRUE);
    nodelay(stdscr, FALSE);
    init_colors();
    init_mouse();

	block_signals();
    edit(argv + 1, 1);
	

    clear();
	refresh();

	clstr = tgetstr("te", NULL);
	if(clstr) putp(clstr);
    endwin();

	save_setup();
	shut_mouse();

    return 0;
}


