/*

   Copyright (C) 1996, 1997 the Free Software Foundation

   Authors: 2000 Steve Grundell

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#include "config.h"
#include "edit.h"
#include "editcmddef.h"
#ifdef SCO_FLAVOR
#       include <sys/timeb.h>
#endif							/* SCO_FLAVOR */
#include <time.h>				/* for ctime() */
#ifdef _WIN32
#    include <io.h>
#    include <fcntl.h>
#else
# include <regex.h>
#endif
#include <ctype.h>

static int get_filetype(WEdit * edit, FILE * f);
static int check_filetype(WEdit * edit, char *file, char *text, char f, char t);

int edit_get_filetype(WEdit * edit)
{
	/* search the user's filetypes file, then the system-wide one */
	if (!get_filetype(edit, open_user_config_file(FILETYPES_FILE)))
		if (!get_filetype(edit, open_system_config_file(FILETYPES_FILE)))
			return 0;

//	if (edit->rc_file)
//		read_setup(edit->rc_file);

	if (edit->syntax_file)
		edit_load_syntax(edit, 0, 0);
	return 1;
}
#define FREE(P) {if(P){free(P);(P)=NULL;}}
#define my_strdup(P) strdup((P==NULL) ? "" : P)
static int get_filetype(WEdit * edit, FILE * f)
{
	int line = 0, l, nl, more;
	char buf[256], *s, *args, ff = 0, t = 0, *moo = NULL;
	char *name, *file, *text, *syntax, *command, *help, *rc;

	if (!f)
		return 0;

	name = file = text = syntax = command = help = rc = NULL;
	FREE(edit->ft_name);
	FREE(edit->ft_filename);
	FREE(edit->ft_text);
	FREE(edit->ft_command);
	FREE(edit->syntax_file);
	FREE(edit->rc_file);

	more = 0;
	while (fgets(buf, 255, f)) {
		s = buf;
		line++;
		
		/* leading space, */
		while (*s && isspace(*s)) {
			s = s + 1;
			l--;
		}

		if (more)
			s = catstrs(moo, " ", s, 0);

		l = strlen(s);

		/* chop off newline, */
		if ((nl = (s[l - 1] == '\n')))
			s[--l] = 0;

		/* get the rest of the line if no newline found, or if esaped */
		if (!nl || ((s[l - 1] == '\\') && (s[l - 2] != '\\'))) {
			s[--l] = 0;
			more = 1;
			moo = catstrs(s, 0);
			continue;
		} else
			more = 0;

		/* trailing space, */
		while (s[l - 1] == ' ')
			s[--l] = 0;


		if ((*s == 0) || (*s == '\n') || (*s == '\r') || (*s == '#'))
			continue;

		if ((args = strchr(s, ' ')) != NULL) {
			*args = 0;
			args = args + 1;
			while (*args && isspace(*args))
				args = args + 1;
			l = strlen(args);
			if ((args[0] == '"') && (args[l - 1] == '"')) {
				args[l - 1] = 0;
				args = args + 1;
			}
		}

		if (!strcmp(s, "filetype")) {
			if (name) {
				if (check_filetype(edit, file, text, ff, t)) {
					fclose(f);
					edit->ft_name = my_strdup(name);
					edit->ft_filename = my_strdup(file);
					edit->ft_text = my_strdup(text);
					edit->ft_command = my_strdup(command);
					edit->ft_help = my_strdup(help);
					edit->syntax_file = my_strdup(syntax);
					edit->rc_file = my_strdup(rc);
					
					return 1;
				}
			}
			name = catstrs(args, 0);
			file = text = syntax = command = help = rc = NULL;
			ff = t = 0;
		} else if (!strcmp(s, "file") || !strcmp(s, "+file")) {
			file = catstrs(args, 0);
			ff = *s;
		} else if (!strcmp(s, "text") || !strcmp(s, "+text")) {
			text = catstrs(args, 0);
			t = *s;
		} else if (!strcmp(s, "syntax"))
			syntax = catstrs(args, 0);
		else if (!strcmp(s, "command"))
			command = catstrs(args, 0);
		else if (!strcmp(s, "help"))
			help = catstrs(args, 0);
		else if (!strcmp(s, "rc"))
			rc = catstrs(args, 0);
		
	}
	fclose(f);

	if (name)
	{
		if (check_filetype(edit, file, text, ff, t)) {
			edit->ft_name = my_strdup(name);
			edit->ft_filename = my_strdup(file);
			edit->ft_text = my_strdup(text);
			edit->ft_command = my_strdup(command);
			edit->ft_help = my_strdup(help);
			edit->syntax_file = my_strdup(syntax);
			edit->rc_file = my_strdup(rc);
			return 1;
		}
	}
	return 0;
}

static int check_filetype(WEdit * edit, char *file, char *text, char ff, char t)
{
#ifndef _WIN32
	regex_t r;
	regmatch_t pmatch[1];
	int m = 0, i;
	char buf[256], *fn;
	FILE *f;

	if (*edit->filename && file) {
		if (regcomp(&r, file, REG_EXTENDED))
			return 0;
		if ((*edit->dir) || (*edit->filename == '/'))
			fn = concat_dir_and_file(edit->dir, edit->filename);
		else {
			fn = concat_dir_and_file(working_dir(), edit->filename);
		}
		m = regexec(&r, fn, 1, pmatch, 0);
		regfree(&r);
		if (m == 0) {
			if (t != '+')
				return 1;
		} else {
			if (ff == '+')
				return 0;
		}
	}

	if (!text)
		return 0;

	if (!(f = fopen(edit->filename, "r")))
		return ((m == 0) && (t != '+'));

	for (i = 0; i < 2; i++) {
		if (!fgets(buf, 255, f)) {
			fclose(f);
			return 0;
		}
		m = strlen(buf);
		if (buf[m - 1] == '\n')
			buf[m - 1] = 0;

		if (regcomp(&r, text, REG_EXTENDED)) {
			fclose(f);
			return 0;
		}
		m = regexec(&r, buf, 1, pmatch, 0);
		regfree(&r);
		if (m == 0) {
			fclose(f);
			return 1;
		}
	}
	fclose(f);
#endif
	return 0;
}
