/* 
   Copyright (C) 2003 the Free Software Foundation

   Authors: 2003 Mateusz Golicz <mteg@jaszczur.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#include "config.h"
#include "edit.h"
#include <sys/param.h>
#include <unistd.h>


extern WEdit * wedit;
extern Dlg_head * edit_dlg;
WEdit * wedit_current;
WEdit * wedit_mark = NULL;
int wedit_count = 1;
static Widget * wedit_filebar;

void multifile_do_fullscreen(WEdit * we);

void multifile_cleanup(void)
{
	WEdit * e, * p;
	
	for(p = wedit, e = wedit->next; e; p = e, e = e->next)
	{
		if(!e->running)
		{
			if(wedit_current == e) wedit_current = e->next;
			if(wedit_mark == e) wedit_mark = e->next;
			p->next = e->next;	
			destroy_widget ((Widget*) e);
		}
	}
	
	if(! wedit->running)
	{
		e = wedit;
		if(wedit_current == e) wedit_current = e->next;
		if(wedit_mark == e) wedit_mark = e->next;
		wedit = wedit->next;
		
		destroy_widget ((Widget*) e);
	}
	if(!wedit_current) wedit_current = wedit;
	if(!wedit_mark) wedit_mark = wedit;
}

int multifile_filebar_callback (Dlg_head *h, Widget *l, int Msg, int Par)
{
	if(Msg == WIDGET_INIT) return(1);
	if(Msg == WIDGET_FOCUS) return(0);
	if(Msg == WIDGET_DRAW && wedit_count > 1)
	{
		int start = 0, count = 0, num = -1, xspace = COLS + 1, c = 0, tc;
		char format[20];
		WEdit * e;
		
//		fprintf(stderr, "Begin redraw!\n");

		while(num == -1)
		{
			xspace = 0; count = 0; num = -1; c = 0;
			
			for(e = wedit; e; e = e->next, c++)
			{
				if(c >= start)
				{
					char * name = rindex(e->filename, '/');
					int current;
					
					if(name)
						name++;
					else
						name = e->filename;
					
					if(name)
					{
						current = strlen(name) + 5;
						if((current * 3) > COLS)
							current = COLS / 3;
					}
					else
						current = 6;
					
					if((xspace + current) > COLS) break;
					xspace += current;
					
					if(e == wedit_current) num = count;
					count ++;
				}
			}
//			fprintf(stderr, "For start: %d xspace is %d and count is %d and num is %d\n", start, xspace , count, num);
			start++;
			if(num == (count - 1) && e)
				num = -1;
		}

		start--;
//		fprintf(stderr, "Agreed to that (S:%d)\n", start);
		snprintf(format, 20, "[%%2d:%%.%ds]", COLS / 3 - 5);
		widget_move(l, 0, 0);

		for(e = wedit, c = 0, tc = 0; e; e = e->next)
		{
			tc++;
			if(start)
				start--;
			else
			{
				char * name = rindex(e->filename, '/');
				
				c++;
				if(c > count) break;
				
				if(name)
					name++;
				else
					name = e->filename;
				
				if(!name) name = "?";

				if(e == wedit_current)
					attrset(ERROR_COLOR);
				else
				{
					if(e->modified)
						attrset(MARKED_SELECTED_COLOR);
					else
						attrset(SELECTED_COLOR);
				}
				
//				fprintf(stderr, "Printw: format: %s arg : %d, %s\n", format, tc, name);
				printw(format, tc, name);

			}
		}
		
		attrset(SELECTED_COLOR);
		xspace = COLS - xspace;
		while(xspace--) printw(" ");
		
		return 1;
	}
	return default_proc (h, Msg, Par);
}

int multifile_filebar_destroy (Widget *w)
{
	return(1);
}


void multifile_init(WEdit * we)
{

	wedit_filebar = xmalloc(sizeof (Widget), "multifile_init");

	wedit_current = we;
	wedit_mark = we;
	we->next = NULL;
	we->running = 1;
	we->fullscreen = 0;

        init_widget (wedit_filebar, LINES - 2, 0, 1, 1,
	                (callback_fn) multifile_filebar_callback,
			(destroy_fn) multifile_filebar_destroy, NULL, "fbar");

	add_widget(we->widget.parent, wedit_filebar);
}

int edit_mouse_event (Gpm_Event * event, void *x);
static int edit_callback (Dlg_head * h, WEdit * edit, int msg, int par);

void multifile_renew(void)
{
	WEdit * e;
	for(e = wedit; e; e = e->next)
		edit_scroll_screen_over_cursor (e);
}

void multifile_nofullscreen(void)
{
	WEdit * e;
	for(e = wedit; e; e = e->next)
		if(e->fullscreen) multifile_do_fullscreen(e);
}

WEdit * multifile_new(char * file, int append)
{
	int inlines, cline, dy;
	WEdit * newedit;
	char * text = NULL;
	int again = 0;
	int bgain = 0;
	int needed = 0, l;
	WEdit * pred;
	
	multifile_nofullscreen();

	if(append)
		for(pred = wedit; pred->next; pred = pred->next);
	else
		pred = wedit_current;
	
	inlines = pred->widget.lines / 2;
	if(inlines < 2) needed = 2;
	
	if(needed)
	{
		WEdit * e;
		/* windows above pred */
		for(e = wedit; e && e != pred; e = e->next)
		{
			e->widget.y -= again;
			if(e->widget.lines > 2 && needed)
			{
				int localgain;
				localgain = (e->widget.lines - 2);
				if(localgain > needed) localgain = needed;
				again += localgain;
				needed -= localgain;
				e->widget.lines -= localgain;
				e->num_widget_lines -= localgain;
			}
		}
		if(needed) return(NULL);
		
		pred->widget.y -= again;
		pred->widget.lines += again + bgain;
		pred->num_widget_lines += again + bgain;

		inlines = again + bgain;
	}
	
	if(file) 
	{
		if (!(*file))
		{
		file = NULL;
		text = "";
		}
	}
	else
		text = "";

	if (!(newedit = edit_init (NULL, inlines - 1, COLS, file, text, "", 0))) {
		message (1, _(" Error "), get_error_msg (""));
		return 0;
	}
	newedit->macro_i = -1;

	dy = pred->widget.lines - inlines;

	pred->widget.lines = dy;
	pred->num_widget_lines = dy - 1;

//	fprintf(stderr, "CREAT: dy = %d, inlines = %d\n", dy, inlines);

	init_widget (&(newedit->widget), pred->widget.y + dy, 0, inlines, COLS,
				 (callback_fn) edit_callback,
				 (destroy_fn) edit_clean,
				 (mouse_h) edit_mouse_event, 0);
	if(wedit_count == 1)
	{
		newedit->widget.lines -= 1;
		newedit->num_widget_lines -= 1;
	}

//	fprintf(stderr, "SUMUP: old: %d:%d new: %d:%d\n", pred->widget.y, pred->widget.lines, newedit->widget.y, newedit->widget.lines);

	widget_want_cursor (newedit->widget, 1);

	cline = storeline_get(file);
	edit_move_display (newedit, ((l = cline - 1 - (newedit->num_widget_lines / 2)) > 0) ? l : 0);
	edit_move_to_line (newedit, cline - 1);
	newedit->force |= REDRAW_COMPLETELY;

	add_widget (edit_dlg, newedit);

	newedit->running = 1;


	newedit->next = pred->next;
	pred->next = newedit;


	wedit_current = newedit;

	if(!option_auto_fullscreen)
		multifile_renew();
	
	wedit_count ++;

	multifile_filebar_callback(wedit_filebar->parent, wedit_filebar, WIDGET_DRAW, 0);

	wedit_current->fullscreen = 0;
	if(option_auto_fullscreen)
		multifile_fullscreen(wedit_current);

	if(again || bgain)
	{
	        do_refresh();
	        refresh();
	        doupdate();
	}

//	fprintf(stderr, "FULLSCREEN SUMUP: old: %d:%d new: %d:%d\n", pred->widget.y, pred->widget.lines, newedit->widget.y, newedit->widget.lines);

	return newedit;
}


static int edit_callback (Dlg_head * h, WEdit * e, int msg, int par)
{
	switch (msg) {
	case WIDGET_INIT:
		e->force |= REDRAW_COMPLETELY;
		edit_labels (e);
		break;
	case WIDGET_DRAW:
		e->force |= REDRAW_COMPLETELY;
		e->num_widget_columns = COLS;
	case WIDGET_FOCUS:
		edit_update_screen (e);
		return 1;
	case WIDGET_KEY:{
			int cmd, ch;
		e = wedit_current;
//		fprintf(stderr, "%x\n", par);
			if (edit_drop_hotkey_menu (e, par))		 /* first check alt-f, alt-e, alt-s, etc for
drop menus */
				return 1;
			if (!edit_translate_key (e, 0, par, get_key_state (), &cmd, &ch))
				return 0;
			edit_execute_key_command (e, cmd, ch);
			edit_update_screen (e);
		}
		return 1;
	case WIDGET_COMMAND:
		edit_execute_key_command (e, par, -1);
		edit_update_screen (e);
		return 1;
	case WIDGET_CURSOR:
		widget_move (&e->widget, e->curs_row + EDIT_TEXT_VERTICAL_OFFSET, e->curs_col + e->start_col);
		return 1;
	}
	return default_proc (h, msg, par);
}

int (*multifile_getcallback(void))(Dlg_head * h, WEdit * e, int msg, int par)
{
	return(edit_callback);
}

void multifile_pass_fullscreen(WEdit * prev, WEdit * next, int fs)
{
	if(fs)
	{
		if(prev->fullscreen)
		{
			prev->num_widget_lines = prev->had_ysize - 1;
			prev->widget.lines = 0;
			prev->widget.y = prev->had_pos;
			prev->fullscreen = 0;
			next->num_widget_lines = next->widget.parent->lines - 2 - 1;
			next->widget.lines = next->widget.parent->lines - 1 - 1;
			next->had_pos = next->widget.y;
			next->widget.y = 0;
			next->fullscreen = 1;
			do_refresh();
			refresh(); 
			doupdate();
		}
		else
			multifile_fullscreen(next);
		
	}
	else
		if(prev->fullscreen) multifile_fullscreen(prev);

}


void multifile_next(int fs)
{
	WEdit * next = wedit_current->next;
	if(!next) next = wedit;
	
	if(next == wedit_current) return;
	
	dlg_select_widget (wedit_current->widget.parent, next);
	multifile_pass_fullscreen(wedit_current, next, fs);
	wedit_current = next;
	wedit_current->force |= REDRAW_PAGE;
	multifile_filebar_callback(wedit_filebar->parent, wedit_filebar, WIDGET_DRAW, 0);
}


void multifile_prev(int fs)
{
	WEdit * prev = wedit;
	
	
	if(wedit == wedit_current)
		for(prev = wedit; prev->next; prev = prev->next);
	else
		for(; prev; prev = prev->next)
			if(prev->next == wedit_current) break;
	
	if(!prev) return;

	if(prev == wedit_current) return;

	dlg_select_widget (wedit_current->widget.parent, prev);
	multifile_pass_fullscreen(wedit_current, prev, fs);
	wedit_current = prev;
	wedit_current->force |= REDRAW_PAGE;
	multifile_filebar_callback(wedit_filebar->parent, wedit_filebar, WIDGET_DRAW, 0);
}


void multifile_switchto(WEdit *e, int fs)
{
	if(e == wedit_current) return;

	dlg_select_widget (wedit_current->widget.parent, e);
	multifile_pass_fullscreen(wedit_current, e, fs);
	wedit_current = e;
	wedit_current->force |= REDRAW_COMPLETELY;
	edit_update_screen(wedit_current);

	multifile_filebar_callback(wedit_filebar->parent, wedit_filebar, WIDGET_DRAW, 0);

}

void multifile_switch(int n, int fs)
{
	WEdit * e = wedit;
	
	while(e && (n--))
		e = e->next;
	
	if(!e) return;
	multifile_switchto(e, fs);
}

void multifile_select(WEdit * we)
{
	dlg_select_widget (wedit_current->widget.parent, we);
	wedit_current = we;
}

void multifile_clean_markers(WEdit * we)
{
	if(we != wedit_mark)
	{
		if(wedit_mark)
		{
			edit_set_markers(wedit_mark, 0, 0, 0, 0);
			wedit_mark->force |= REDRAW_PAGE;
			edit_refresh_cmd(wedit_mark);
		}
		wedit_mark = we;
	}
}

void multifile_resize(WEdit * we, int nsize)
{
	if(we->fullscreen) return;
	switch(nsize)
	{
		case 1:
		default:
			{
				WEdit * e;
				
				for(e = we->next; e; e = e->next)
					if(e->widget.lines > 2) break;
				if(e)
				{
						we->widget.lines ++;
						we->num_widget_lines ++;	
					we->force |= REDRAW_COMPLETELY;		
					for(e = we->next; e; e = e->next)
					{
						e->force |= REDRAW_COMPLETELY;
						e->widget.y ++;
						if(e->widget.lines > 2)
						{
							e->widget.lines --;
							e->num_widget_lines --;
							break;
						}
					}
				}
			}
			break;
		case -1:
			if(we->next && we->widget.lines > 2)
			{
					we->widget.lines --;
					we->num_widget_lines --;
				we->next->widget.y --;
					we->next->widget.lines ++;
					we->next->num_widget_lines ++;
				we->force |= REDRAW_COMPLETELY;
				we->next->force |= REDRAW_COMPLETELY;
			}
			break;
	}
	if(we->force & REDRAW_COMPLETELY)
	{
		multifile_renew();
		do_refresh();
		refresh();
	}
}

void multifile_do_fullscreen(WEdit * we)
{
	if(we->fullscreen)
	{
		WEdit * e;
		for(e = wedit; e; e = e->next)
			if(! e->fullscreen)
				e->widget.lines = e->had_ysize;

		we->num_widget_lines = we->had_ysize - 1;
		we->widget.lines = we->had_ysize;
		we->widget.y = we->had_pos;
		we->fullscreen = 0;
	}
	else
	{
		WEdit * e;
		if(wedit_count == 1) return;
		for(e = wedit; e; e = e->next)
		{
			if(e == we) continue;
			e->had_ysize = e->widget.lines;
			e->widget.lines = 0;
		}

		we->had_ysize = we->widget.lines;
		we->had_pos = we->widget.y;
		
		we->num_widget_lines = we->widget.parent->lines - 2 - 1;
		we->widget.lines = we->widget.parent->lines - 1 - 1;
		we->widget.y = 0;
		we->fullscreen = 1;
	}
	multifile_renew();
}

void multifile_fullscreen(WEdit * we)
{
	multifile_do_fullscreen(we);
	do_refresh();
	refresh(); 
	doupdate();
}

int multifile_quit(WEdit * we)
{
	Dlg_head * h;
	WEdit * next;
	WEdit * prev;
	int bonus = (wedit_count == 2) ? 1 : 0;
	int was_fs = 0;
	
//	fprintf(stderr, "QUIT()\n");
	
	if(we->fullscreen)
	{
		multifile_fullscreen(we);
		was_fs = 1;
	}

//	fprintf(stderr, "QUIT() 1\n");
	
	next = we->next;
	if(!next) next = wedit;
	if(next == we)
	{
		h = we->widget.parent;
		h->running = 0;
		wedit_count --;
		remove_widget(h, we);
//		fprintf(stderr, "QUIT() DONE\n");
		return(1);
	}

//	fprintf(stderr, "QUIT() 2\n");

	
	for(prev = wedit; prev; prev = prev->next)
		if(prev->next == we) break;

//	fprintf(stderr, "QUIT() 3\n");

	if(prev)
	{
//		fprintf(stderr, "QUIT() 4\n");

		prev->num_widget_lines += we->widget.lines + bonus;
		prev->widget.lines += we->widget.lines + bonus;
	}
	else
	{
//		fprintf(stderr, "QUIT() 5\n");

		next->widget.y -= we->widget.lines;
		next->widget.lines += we->widget.lines + bonus;
		next->num_widget_lines += we->widget.lines + bonus;
	}
	
//	fprintf(stderr, "QUIT() 6\n");

	h = we->widget.parent;
	h->running = 0;

	remove_widget(h, we);
	we->running = 0;
	
//	if(we == wedit)
//	{
//		next->next = wedit->next;
//		wedit = next;
//		wedit->next = next->next;
//		next->next = wedit;
//		wedit = next;
//	}
	
//	fprintf(stderr, "QUIT() 7\n");

	for(; ! next->running; next = next->next);
	if(next) multifile_select(next);

//	fprintf(stderr, "QUIT() 8\n");

	next->force |= REDRAW_COMPLETELY;

//	fprintf(stderr, "QUIT() HALFWAY\n");

	
	multifile_nofullscreen();

	
	wedit_count --;

	if(was_fs && wedit_count > 0)
		multifile_fullscreen(wedit_current);

	multifile_filebar_callback(wedit_filebar->parent, wedit_filebar, WIDGET_DRAW, 0);
	clear();
	do_refresh();
	refresh(); 
	doupdate();
//	fprintf(stderr, "QUIT() DONE\n");

	return(1);
}

int multifile_count(void)
{
	return(wedit_count);
}

void multifile_shutdown(void)
{
	WEdit * e, * caller, *n;
/*	int m = 0;

	for(e = wedit; e; e = e->next)
		if(e->modified) m = 1;

	if(m)
		if(!edit_query_dialog2 (_(" Quit "), _(" Some of the files were modified. Really quit? "), _(" &No "), _(" &Yes "))) return;

	wedit_count = 0;
	wedit->widget.parent->running = 0;*/
	
	caller = wedit_current;
	
//	for(e = wedit; e; e = e->next)
//		fprintf(stderr, "Order %x (n = %x)\n", e, e->next);
	
	for(e = wedit; e; e = n)
	{
//		fprintf(stderr, "Loop: %x (N = %x)\n", e, e -> next);
		n = e->next;
		if(e != caller)
		{
//			fprintf(stderr, "SwitchTo\n");
			multifile_switchto(e, option_auto_fullscreen);
//			fprintf(stderr, "SendQuit\n");
			if(!edit_quit_cmd(e, 0))
			{
//				fprintf(stderr, "Breakage\n");
				break;
			}
//			fprintf(stderr, "No breakage\n");
		}
	}
	
//	fprintf(stderr, "Finish: %d\n", e);
	
	if(e)
		return;

	edit_quit_cmd(caller, 1);
	wedit_count = 0;
	wedit->widget.parent->running = 0;
}

void multifile_filebar_refresh(void)
{
        multifile_filebar_callback(wedit_filebar->parent, wedit_filebar, WIDGET_DRAW, 0);
}

WEdit * multifile_findfile(char * file)
{
	char path1[MAXPATHLEN];
	char path2[MAXPATHLEN];
	WEdit * e;
	
	if(!realpath(file, path1)) return(NULL);
	
	for(e = wedit; e; e = e->next)
	{
		if(realpath(e->filename, path2))
			if(strcmp(path1, path2) == 0) return(e);
	}
	return(NULL);
}

