/* setup options

   Copyright (C) 2000 Steve Grundell
   Copyright (C) 1994 Miguel de Icaza.
   Copyright (C) 2003 Mateusz Golicz <mteg@jaszczur.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <ctype.h>
#include "config.h"
#include "edit.h"

typedef struct {
	char *name;
	char type;					/* i = int, s = string */
	void *var;
} setting;
static setting settings[] = {
	{"word_wrap_line_length", 'i', &option_word_wrap_line_length},
	{"typewriter_wrap", 'i', &option_typewriter_wrap},
	{"auto_para_formatting", 'i', &option_auto_para_formatting},
	{"international_characters", 'i', &option_international_characters},
	{"tab_spacing", 'i', &option_tab_spacing},
	{"fill_tabs_with_spaces", 'i', &option_fill_tabs_with_spaces},
	{"return_does_auto_indent", 'i', &option_return_does_auto_indent},
	{"backspace_through_tabs", 'i', &option_backspace_through_tabs},
	{"fake_half_tabs", 'i', &option_fake_half_tabs},
	{"save_mode", 'i', &option_save_mode},
	{"backup_ext_int", 'i', &option_backup_ext_int},
	{"find_bracket", 'i', &option_find_bracket},
	{"max_undo", 'i', &option_max_undo},
	{"eight_bit_clean", 'i', &option_eight_bit_clean},
	{"auto_fullscreen", 'i', &option_auto_fullscreen},
	{"store_filepos", 'i', &option_storeline},
	{"edit_right_extreme", 'i', &option_edit_right_extreme},
	{"edit_left_extreme", 'i', &option_edit_left_extreme},
	{"edit_top_extreme", 'i', &option_edit_top_extreme},
	{"edit_bottom_extreme", 'i', &option_edit_bottom_extreme},
	{"whole_chars_search", 's', &option_whole_chars_search},
	{"whole_chars_move", 's', &option_whole_chars_move},
	{"backup_ext", 's', &option_backup_ext},
	{0, 0, 0}
};


void init_settings();
void read_setup_file(FILE * f);

#ifdef DEBUG
void dump_setup() {
	int i = 0;
	while (settings[i].type) {
		switch (settings[i].type) {
		case 'i':
			fprintf(stderr, "%s %i\n", settings[i].name,*((int *)settings[i].var));
			break;
		case 'S':
		case 's':
			fprintf(stderr, "%s %s\n", settings[i].name,*((char **)settings[i].var));
			break;
		}
		i++;
	}
}
#endif
void read_setup(const char *file)
{
	if (!file || !*file)
		file = RC_FILE;
	read_setup_file(open_system_config_file(file));
	read_setup_file(open_user_config_file(file));
	read_setup_file(fopen(file, "r"));
}

void read_setup_file(FILE * f)
{
	char buf[256], *s, *arg;
	int line = 0, l, i;

	if (!f)
		return;

	while (fgets(buf, 255, f)) {
		line++;
		s = buf;
		l = strlen(s);
		/* chop off newline, */
		if (s[l - 1] == '\n')
			s[--l] = 0;
		/* trailing space, */
		while (s[l - 1] == ' ')
			s[--l] = 0;
		/* leading space, */
		while (*s && isspace(*s))
			s = s + 1;
		if ((*s == 0) || (*s == '\n') || (*s == '\r') || (*s == '#'))
			continue;

		if (!(arg = strchr(s, ' ')))
			continue;

		*arg = 0;
		do
			arg = arg + 1;
		while (*arg && isspace(*(arg)));
		i = 0;
		
		while (settings[i].type) {
			if (!strcmp(s, settings[i].name)) {
				switch (settings[i].type) {
				case 'i':
					*((int *)settings[i].var) = (int)atoi(arg);
					break;
				case 'S':
					if (settings[i].var)
						free(*(char **)settings[i].var);
				case 's':
					settings[i].type = 'S';
					*((char **)settings[i].var) = strdup(arg);
					break;
				}
				break;
			}
			i++;
		}
	}
	fclose(f);

}

void save_setup(void)
{
	FILE *f;
	int i = 0;
	
	f = fopen(concat_dir_and_file(conf_dir, RC_FILE), "w");
	if(f)
	{
		while (settings[i].type) {
			switch (settings[i].type) {
				case 'i':
					fprintf(f, "%s %d\n", settings[i].name, *((int *)settings[i].var));
					break;
				case 's':
				case 'S':
					fprintf(f, "%s %s\n", settings[i].name, *((char **)settings[i].var));
					break;
			}
			i++;
		}
	}
	fclose(f);
}
