/* store current line

   Copyright (C) 2003 Free Software Foundation
   Autors: 2003 Mateusz Golicz <mteg@jaszczur.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <ctype.h>
#include <stdio.h>
#include <sys/param.h>
#include <unistd.h>
#include <string.h>
#include "edit.h"
#include "editcmddef.h"
#include "config.h"

void storeline_save(const char * filename, int numline)
{
	unsigned char * lfile = catstrs(conf_dir, "/cooledit.filepos", 0);
	unsigned char * tempfile = catstrs(conf_dir, "/cooledit.filepos.tmp", 0);
	unsigned char path[MAXPATHLEN];
	FILE * in;
	FILE * out;
	int w = 1;
	
	if(!option_storeline) return;
	if(!filename) return;
	if(!*filename) return;
	
	if(!realpath(filename, path)) return;
	if(index(path, '\n')) return;
	
	in = fopen(lfile, "r");
	if(!(out = fopen(tempfile, "w")))
	{
		if(in) fclose(in);
		return;
	}
	
	if(in)
	{
		unsigned char line[MAXPATHLEN + 12];
		unsigned char * s;
		
		while((s = fgets(line, MAXPATHLEN + 12, in)))
		{
			unsigned char * n = index(line, '\n');
			if(n) *n = 0;
			
			n = index(line, ' ');
			*n = 0;
			n++;
			
			if(!strcmp(n, path))
			{
				if(w) fprintf(out, "%d %s\n", numline, path);
				w = 0;
			}
			else
				fprintf(out, "%s %s\n", s, n);
			
		}
		fclose(in);
	}

	if(w) fprintf(out, "%d %s\n", numline, path);

	fclose(out);
	
	if(rename(tempfile, lfile) == -1) perror("rename");
}

int storeline_get(const char * filename)
{
	unsigned char * lfile = catstrs(conf_dir, "/cooledit.filepos", 0);
	unsigned char path[MAXPATHLEN];
	unsigned char line[MAXPATHLEN + 12];
	FILE * in;
	unsigned char * s;
	
	if(!option_storeline) return(1);
	if(!filename) return(1);
	if(!*filename) return(1);
	if(!realpath(filename, path)) return(1);
	if(index(path, '\n')) return(1);
	
	in = fopen(lfile, "r");
	if(!in) return(1);
	
	while((s = fgets(line, MAXPATHLEN + 12, in)))
	{
		unsigned char * n = index(line, '\n');
		if(n) *n = 0;
			
		n = index(line, ' ');
		*n = 0;
		n++;
			
		if(!strcmp(n, path))
		{
			fclose(in);
			return(atoi(s));
		}
	}
	fclose(in);
	return(1);
}
